#!/usr/bin/env python3

import numpy as np
import sys
from argparse import ArgumentParser, FileType
from pathlib import Path

this_dir = Path(__file__).parent


def lonlat(string):
    pair = string.split(",")
    return (float(pair[0]), float(pair[1]))


parser = ArgumentParser()
parser.add_argument("--descriptorfile", type=FileType('r'), default=this_dir.parent/"descriptorfiles"/"icon_1mom.py", help="")
parser.add_argument("--coords", type=lonlat, nargs="+", required=True, help="Coordinates of the groundstations")
parser.add_argument("--frequency", type=float, nargs="+", default=94)
parser.add_argument("--dt", type=str, default=None, help="Timestep")
parser.add_argument("--source_component", type=str, default="atmo_output", help="component name of the source component")
parser.add_argument("--source_grid", type=str, default="icon_atmos_grid", help="grid name of the source grid")
args = parser.parse_args()


from yac import YAC, UnstructuredGrid, Location, Field, TimeUnit, \
    InterpolationStack, NNNReductionType, Reduction, Action
import pyPamtra


yac = YAC()

comp = yac.def_comp("pamtra")

coords = np.deg2rad(args.coords)
ncoords = len(args.coords)
grid = UnstructuredGrid("pamtra-grid", [3]*ncoords,
                        [*coords[:, 0], *(coords[:, 0]+0.1), *coords[:, 0]],
                        [*coords[:, 1], *coords[:, 1], *(coords[:, 1]+0.1)],
                        np.array(list(zip(range(ncoords), range(ncoords, 2*ncoords), range(2*ncoords, 3*ncoords)))).ravel())
points = grid.def_points(Location.CELL,
                         coords[:, 0], coords[:, 1])

yac.sync_def()

# print(yac.get_field_names(*source), file=sys.stderr)

interp = InterpolationStack()
interp.add_nnn(NNNReductionType.AVG, 1, 1.0)

source = (args.source_component, args.source_grid)

dt = args.dt or yac.get_field_timestep(*source, "temp")

fields = {}
for name in ["temp", "pres", "qv", "qi", "qc", "qs", "qr", "qg", "z_ifc", "u", "v", "w"]:
    nlevel = yac.get_field_collection_size(*source, name)
    fields[name] = Field.create(name, comp, points, nlevel,
                                dt, TimeUnit.ISO_FORMAT)
    yac.def_couple(*source, name,
                   "pamtra", "pamtra-grid", name,
                   dt, TimeUnit.ISO_FORMAT, Reduction.TIME_NONE,
                   interp)

yac.enddef()

with open(args.descriptorfile, "r") as dfile:
    exec(dfile.read())  # adds a variable "descriptorFile"


def call_pamtra(temp, pres, qv, qi, qc, qs, qr, qg, z_ifc, u, v, w,
                datetime):
    pamtra = pyPamtra.pyPamtra()
    for hydro in descriptorFile:
        pamtra.df.addHydrometeor(hydro)
    pamtra.nmlSet["passive"] = False
    pamtra.nmlSet["active"] = True
    pamtra.nmlSet["radar_mode"] = "simple"
    rh = 100*pyPamtra.meteoSI.q2rh(qv, temp, pres)
    pamtra.createProfile(hgt_lev=z_ifc.T[None, :, ::-1],
                         press=pres.T[None, :, ::-1],
                         temp=temp.T[None, :, ::-1],
                         relhum=rh.T[None, :, ::-1],
                         hydro_q=np.stack([qc.T, qi.T, qr.T, qs.T, qg.T], axis=-1)[:, ::-1, :],
                         )
    pamtra.runPamtra(args.frequency)
    #pamtra.writeResultsToNetCDF(f"output_{datetime}.nc", xarrayCompatibleOutput=True)
    return pamtra # return the entire pamtra object and let the caller decide what to take as results

# prepare output
import xarray as xr
import pandas as pd
Ntimesperoutput = 6 # TODO should be user config...
pam_time_idx = xr.IndexVariable(
        dims='time_idx',
        data=np.arange(0, Ntimesperoutput),
        attrs={'name':'time_idx',
               'long_name':'index of the time step within the output file'})

pam_station_idx = xr.IndexVariable(
        dims='sta_idx',
        data=np.arange(0, ncoords),
        attrs={'name':'sta_idx',
               'long_name':'index of the station',
               'comment':'look at variables sta_coords for location'})

pam_hgt_idx = xr.IndexVariable(
        dims='hgt_idx',
        data=np.arange(0, nlevel-1),
        attrs={'name':'hgt_idx',
               'long_name':'height indx above surface',
               'direction':'bottom-up',
               'comment':'look at variable height to get the height profile above each station'})

#pam_sta_name.xr.DataArray(dims='station') # TODO perhaps one can provide station names as input, or instead of giving coordinates can also ask for certain locations to be read from a catalog
pam_time = xr.DataArray(
        dims=['time_idx'],
        coords={'time_idx':pam_time_idx},
        data=np.arange(0, Ntimesperoutput),
        attrs={'name':'time',
               'long_name':'unixtime nanoseconds since 1970',
               'units':'nanoseconds'})

pam_lat = xr.DataArray(
        dims=['sta_idx'],
        coords={'sta_idx':pam_station_idx},
        data=np.rad2deg(coords[:, 1]),
        attrs={'name':'lat',
               'long_name':'station latitude',
               'unit':'degrees North'})

pam_lon = xr.DataArray(
        dims=['sta_idx'],
        coords={'sta_idx':pam_station_idx},
        data=np.rad2deg(coords[:, 0]),
        attrs={'name':'lon',
               'long_name':'station longitude',
               'unit':'degrees East'})

pam_hgt = xr.DataArray(
        dims=['sta_idx', 'hgt_idx'],
        coords={'sta_idx':pam_station_idx,
                'hgt_idx':pam_hgt_idx},
        data=np.empty((ncoords, nlevel-1)),
        attrs={'name':'hgt',
               'long_name':'height above surface',
               'units':'meters'})

pam_Ze = xr.DataArray(
        dims=['sta_idx', 'hgt_idx', 'time_idx'],
        coords={'sta_idx':pam_station_idx,
                'hgt_idx':pam_hgt_idx,
                'time_idx':pam_time_idx},
        data=np.empty((ncoords, nlevel-1, Ntimesperoutput)),
        attrs={'name':'Ze',
               'long_name':'unattenuated reflectivity',
               'units':'dBZ'})

variables = {'height':pam_hgt,
             'time':pam_time,
             'lat':pam_lat,
             'lon':pam_lon,
             'Ze':pam_Ze}

coordinates = {'time_idx':pam_time_idx,
               'station_idx':pam_station_idx,
               'hgt_idx':pam_hgt_idx}

import os
import socket
from datetime import datetime as datetimelib ### ugly but datetime is reused later and can be uglier
global_attributes = {'created_by':os.environ['USER'],
                     'host_machine':socket.gethostname(),
                     'created_on':str(datetimelib.now()),
                     'comment':'This is a test output for ground based stations'}

dataset = xr.Dataset(data_vars=variables,
                     coords=coordinates,
                     attrs=global_attributes)

itime = 0
while True:
    datetime = fields["temp"].datetime
    data = {}
    for v, f in fields.items():
        data[v], info = f.get()
    pamtra = call_pamtra(**data, datetime=datetime)
    pam_time[itime] = pd.to_datetime(datetime) 
    pam_Ze[:, :, itime] = pamtra.r['Ze'][:,0,:,0,0,0] # this extracts only 1 frequency... need to think about it

    itime += 1
    if itime == Ntimesperoutput:
        pam_hgt[:, :] = pamtra.r['radar_hgt'][:, 0, :]
        dataset.to_netcdf('groundstation_{}.nc'.format(pd.to_datetime(pam_time[0].values).strftime('%Y%m%d-%H%M%S')))
        itime = 0

    if info == Action.GET_FOR_RESTART:
        break

